import os
import json
from typing import Any, Union, Tuple, Sequence

import pandas as pd
import numpy as np
import xarray as xr

from otb.config import ROOT_DIR, DATA_DIR, DATASETS_FP, CACHE_DIR, RETURN_TYPES
from otb.cache import CACHE


class Dataset(object):
    """A singleton helper for in-memory datasets."""

    def __init__(self,
                 name: str,
                 datasets_fp: Union[str, os.PathLike, None] = DATASETS_FP,
                 root_dir: Union[str, os.PathLike, None] = ROOT_DIR,
                 data_dir: Union[str, os.PathLike, None] = DATA_DIR,
                 cache_dir: Union[str, os.PathLike, None] = CACHE_DIR) -> None:
        """Read the currently-supported benchmarking task for loaders and evaluators."""
        self._name = name
        self._datasets_fp = datasets_fp
        self._root_dir = root_dir
        self._data_dir = data_dir
        self._cache_dir = cache_dir
        self._df = self._load_dataset()

    def get_slice(self, start_indices: Sequence[int], end_indices: Sequence[int]) -> pd.DataFrame:
        """Obtain a slice of the underlying dataset from start and end indices."""
        if len(start_indices) == 0 or len(start_indices) != len(end_indices):
            raise ValueError(f"malformed {start_indices}, {end_indices}.")
        ranges = []
        for start_idx, end_idx in zip(start_indices, end_indices):
            if start_idx >= 0 and end_idx <= len(self._df) and start_idx < end_idx:
                ranges.append(np.arange(start_idx, end_idx))
            else:
                raise ValueError(f"requested {start_idx}:{end_idx} out of bounds for df with len {len(self._df)}.")
        included = np.concatenate(ranges)
        return self._df.iloc[included, :].copy(deep=True)

    def get_all(self, data_type: str = "pd", device: str = "") -> Any:
        """Obtain the training data for this dataset from the supplied task."""
        return self._handle_return_type(data=self._df, return_type=data_type)

    def get_train(self, task: dict, data_type: str = "pd") -> Tuple[Any, Any]:
        """Obtain the training data for this dataset from the supplied task."""
        indices = [int(i) for i in task["train_idx"] for i in i.split(":")]
        starts, stops = indices[::2], indices[1::2]
        data = self.get_slice(starts, stops)
        X, y = self._handle_task(data=data, task=task)
        return self._handle_return_type(data=X, return_type=data_type), self._handle_return_type(data=y,
                                                                                                 return_type=data_type)

    def get_test(self, task: dict, data_type: str = "pd") -> Tuple[Any, Any]:
        """Obtain the test data for this dataset from the supplied task."""
        indices = [int(i) for i in task["test_idx"] for i in i.split(":")]
        starts, stops = indices[::2], indices[1::2]
        data = self.get_slice(starts, stops)
        X, y = self._handle_task(data=data, task=task)
        return self._handle_return_type(data=X, return_type=data_type), self._handle_return_type(data=y,
                                                                                                 return_type=data_type)

    def get_val(self, task: dict, data_type: str = "pd") -> Tuple[Any, Any]:
        """Obtain the validation data for this dataset from the supplied task."""
        indices = [int(i) for i in task["val_idx"] for i in i.split(":")]
        starts, stops = indices[::2], indices[1::2]
        data = self.get_slice(starts, stops)
        X, y = self._handle_task(data=data, task=task)
        return self._handle_return_type(data=X, return_type=data_type), self._handle_return_type(data=y,
                                                                                                 return_type=data_type)

    def _handle_task(self, data: pd.DataFrame, task: dict) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Split into features and target, dropping missing and transforming target if needed."""
        if task["dropna"]:
            data = data.dropna()
        X = data[[c for c in data.columns if c not in task["remove"]]]
        y = data[[task["target"]]]
        if task["log_transform"]:
            y = np.log10(y)
        return X, y

    def _handle_return_type(self, data: pd.DataFrame, return_type: str) -> Any:
        """Map the slice of underlying data to the requested type."""
        if return_type not in RETURN_TYPES:
            raise NotImplementedError(f"return type {return_type} not implemented.")
        # switch case
        if return_type == "pd":
            return data
        return getattr(self, f"_convert_to_{return_type}")(data)

    def _convert_to_np(self, data: pd.DataFrame) -> np.ndarray:
        """Map the slice of underlying data to np ndarray."""
        nd_arr = data.to_numpy()
        return nd_arr

    def _convert_to_xr(self, data: pd.DataFrame) -> xr.Dataset:
        """Map the slice of underlying data to xr xarray."""
        ds = data.set_index("time").to_xarray()
        return ds

    def _convert_to_nc(self, data: pd.DataFrame) -> xr.Dataset:
        """Map the slice of underlying data to netCDF (an alias for xr.DataSet)."""
        return self._convert_to_xr(data)

    def _load_dataset(self) -> pd.DataFrame:
        """Load the dataset from cache or disk."""
        if self._name in CACHE:
            return CACHE.get_dataset(self._name)
        else:
            return self._load_dataset_from_disk()

    def _load_dataset_from_disk(self) -> pd.DataFrame:
        """Load the dataset from disk."""
        supported_datasets = self._supported_datasets()
        file_name = supported_datasets[self._name]["local_data_path"]
        fp = os.path.join(self._data_dir, self._name, file_name)
        fp_str = str(fp)
        try:
            file_type = fp_str.split(".")[-1]
        except IndexError:
            raise NotImplementedError(f"unknown or unsupported file type {fp}.")
        # netcdf
        if file_type == "nc":
            ds = xr.load_dataset(fp)
            df = ds.to_dataframe()
        else:
            raise NotImplementedError(f"unknown or unsupported file type {fp}.")

        # update the cache
        CACHE.add_dataset(self._name, df)

        return df

    def _supported_datasets(self) -> dict:
        """Load the datasets configuration file."""
        supported_datasets = json.load(open(self._datasets_fp, 'rb'))

        return supported_datasets



from typing import Union

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


def plot_predictions(y_true: Union[pd.Series, pd.DataFrame], y_pred: Union[pd.Series, pd.DataFrame, np.ndarray],
                     *args) -> None:
    """Plot the actual and predicted values."""
    fig, ax = plt.subplots(figsize=(10, 5))
    fig.autofmt_xdate(rotation=45)

    # change the background color
    plt.rcParams["axes.facecolor"] = "#faf9f6"
    plt.rcParams["font.family"] = "sans serif"

    # add grid lines
    ax.grid(color="gray", linestyle="--", linewidth=0.5, alpha=0.7)

    # add axis labels
    ax.set_xlabel(
        "Time",
        fontsize=15,
    )
    if y_pred[0] < 0:
        ax.set_ylabel(r"$\log_{10} C_n^2$", fontsize=12)
    else:
        ax.set_ylabel(r"$C_n^2$", fontsize=12)

    # plot the actual and predicted values
    ax.scatter(y=y_true.values, x=y_true.index, label="actual", alpha=0.9, c="#00356B", marker="o")
    ax.scatter(y=y_pred, x=y_true.index, label="predicted", alpha=0.9, c="#C90016", marker=".")
    for i in range(len(args)):
        if isinstance(args[i], pd.Series) or (isinstance(args[i], np.ndarray) and len(args[i]) == len(y_true)):
            ax.scatter(y=args[i], x=y_true.index, label="model {}".format(i + 1), alpha=0.9, marker=".")

    # add legend
    ax.legend(
        loc="best",
        fontsize=10,
        frameon=True,
        framealpha=0.5,
    )

    # return the plot
    return fig, ax



from typing import Union

import numpy as np


def apply_fried_height_adjustment(cn2: Union['pd.DataFrame', 'pd.Series', np.ndarray], observed: float,
                                  desired: float) -> Union['pd.DataFrame', 'pd.Series', np.ndarray]:
    """Height adjustment, see Chen 2019 Astron. Astrophys. 19 015.

    Args:
        cn2 (pd.Series): The Cn2 values to adjust.
        desired (float): The desired height.
        observed (float, optional): The observed height. Defaults to 0.0.
    
    Returns:
        pd.Series: The adjusted Cn2 values.
    """
    try:
        assert observed >= 0.0
        assert desired > 0.0
    except AssertionError:
        raise ValueError('The observed and desired heights must be positive.')

    if observed > 0:
        cn2 /= ((observed**(-1 / 3)) * np.exp(-1 * observed / 3200))
        cn2 *= ((desired**(-1 / 3)) * np.exp(-1 * desired / 3200))
    return cn2


def apply_oermann_height_adjustment(
        cn2: Union['pd.DataFrame', 'pd.Series', np.ndarray],
        desired: float,
        observed: float,
        power_law_scaling: float = -4 / 3) -> Union['pd.DataFrame', 'pd.Series', np.ndarray]:
    """Height adjustment, see Oermann 2014, use -2/3 or -4/3 scaling laws.
    
    Args:
        cn2 (pd.Series): The Cn2 values to adjust.
        desired (float): The desired height.
        observed (float, optional): The observed height. Defaults to 15.0.
        power_law_scaling (float, optional): The power law scaling. Defaults to -4/3.

    Returns:
        pd.Series: The adjusted Cn2 values.
    """
    try:
        assert observed > 0.0
        assert desired > 0.0
    except AssertionError:
        raise ValueError('The observed and desired heights must be positive.')

    cn2 *= (desired / observed)**(power_law_scaling)

    return cn2



from typing import Union

import numpy as np


class BaseRegressionModel:

    def __init__(self, name: str, target_name: str, **kwargs):
        self.name = name
        self.target_name = target_name

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        raise NotImplementedError

    def predict(self, X: 'pd.DataFrame'):
        raise NotImplementedError



from typing import Union

import pandas as pd
import numpy as np

from otb.benchmark.models.forecasting.base_model import BaseForecastingModel


class MinuteClimatologyForecastingModel(BaseForecastingModel):
    """A model that predicts the mean value of the target variable for a given time seen during training."""

    def __init__(self,
                 name: str,
                 target_name: str,
                 window_size: int,
                 forecast_horizon: int,
                 time_col_name: Union[str, None] = None,
                 **kwargs):
        super().__init__(name, target_name, window_size, forecast_horizon, **kwargs)
        self.time_col_name = time_col_name if time_col_name is not None else None
        self.means = {}
        self.global_mean = np.nan

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        """Determine the mean value of the target variable seen during training for each time."""
        X = X[[c for c in X.columns if c.startswith(self.target_name)]]

        self.global_mean = np.nanmean(X.values.flatten())

        # compute the mean for each interval in X across all days
        X["time_of_day"] = X.index.time
        X_means = X.groupby("time_of_day").mean()
        # iterate through the rows in X_means
        for i in range(len(X_means)):
            self.means[X_means.index[i]] = np.nanmean(X_means.iloc[i, :].values)

    def predict(self, X: 'pd.DataFrame'):
        """Predict the mean seen during training at the time of day of forecast."""
        times = X.index
        times = pd.to_datetime(times)

        # apply the forecast horizon to each time
        time_step = times[1] - times[0]
        timedelta = pd.Timedelta(seconds=self.forecast_horizon * time_step.total_seconds())

        # add the timedelta to each time
        times = times + timedelta

        # convert to time
        preds = []
        for time in times.time:
            if time in self.means and not np.isnan(self.means[time]):
                preds.append(self.means[time])
            else:
                preds.append(self.global_mean)

        return np.array(preds)



from typing import Union

import pandas as pd
import numpy as np
import lightgbm as lgb

from otb.benchmark.models.regression.base_model import BaseRegressionModel
from otb.benchmark.models.regression.air_water_temperature_difference import AWTModel


class HybridAWTRegressionModel(BaseRegressionModel):
    """A model that uses a hybrid of the AWTModel and a random forest regressor to predict the target."""

    def __init__(self, name: str, target_name: str, time_col_name: Union[str, None] = None, **kwargs):
        super().__init__(name, target_name, **kwargs)
        self.time_col_name = time_col_name if time_col_name is not None else None
        self._lgb_mdl = lgb.LGBMRegressor(**kwargs)
        self._awt_mdl = AWTModel(name=name, target_name=target_name, **kwargs)

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        """Train the underlying LightGBM regressor to form a hybrid model."""
        y_pred_awt = self._awt_mdl.predict(X)
        target = y[self.target_name].values - y_pred_awt.values
        target[np.isnan(y_pred_awt)] = np.nan
        self._lgb_mdl.fit(X, target)

    def predict(self, X: 'pd.DataFrame'):
        """Use generate AWT predictions and correct with trained LightGBM regressor."""
        y_pred_awt = self._awt_mdl.predict(X)
        hybrid_correction = self._lgb_mdl.predict(X)
        preds = y_pred_awt.values + hybrid_correction
        preds[np.isnan(y_pred_awt)] = np.nan
        return preds



from typing import Sequence, Tuple

import numpy as np
import pandas as pd
import sklearn.metrics as sk_m

__all__ = ["r2_score", "root_mean_square_error", "mean_absolute_error", "mean_absolute_percentage_error"]


def is_implemented_metric(metric_name: str) -> bool:
    """Check that the metric is implemented"""
    if metric_name in __all__:
        return True
    return False


def r2_score(y_true: Sequence, y_pred: Sequence) -> Tuple[float, int]:
    """An alias for `sklearn.metrics.r2_score`."""
    y_true, y_pred = _get_valid_indices(y_true=y_true, y_pred=y_pred)
    return _format_metric(float(sk_m.r2_score(y_true=y_true, y_pred=y_pred)), len(y_pred))


def root_mean_square_error(y_true: Sequence, y_pred: Sequence) -> Tuple[float, int]:
    """Calculate RMSE from `sklearn.metrics.mean_squared_error`."""
    y_true, y_pred = _get_valid_indices(y_true=y_true, y_pred=y_pred)
    return _format_metric(float(sk_m.mean_squared_error(y_true=y_true, y_pred=y_pred, squared=False)), len(y_pred))


def mean_absolute_error(y_true: Sequence, y_pred: Sequence) -> Tuple[float, int]:
    """An alias for `sklearn.metrics.r2_score`."""
    y_true, y_pred = _get_valid_indices(y_true=y_true, y_pred=y_pred)
    return _format_metric(float(sk_m.mean_absolute_error(y_true=y_true, y_pred=y_pred)), len(y_pred))


def mean_absolute_percentage_error(y_true: Sequence, y_pred: Sequence) -> Tuple[float, int]:
    """An alias for `sklearn.metrics.r2_score`."""
    y_true, y_pred = _get_valid_indices(y_true=y_true, y_pred=y_pred)
    return _format_metric(float(sk_m.mean_absolute_percentage_error(y_true=y_true, y_pred=y_pred)), len(y_pred))


def _format_metric(metric_value: float, valid_predictions: int) -> dict:
    """Format the metric value and valid predictions into a dict."""
    return {"metric_value": metric_value, "valid_predictions": valid_predictions}


def _get_valid_indices(y_true: Sequence, y_pred: Sequence) -> Tuple[Sequence, Sequence]:
    """Get the valid indices for the supplied sequences."""
    if len(y_true) != len(y_pred):
        raise ValueError(f"y_true and y_pred must have the same length, got {len(y_true)} and {len(y_pred)}")

    y_true = y_true.to_numpy().squeeze()

    # ensure we have numpy arrays in y_pred
    if isinstance(y_true, pd.DataFrame):
        y_pred = y_pred.values
    elif isinstance(y_true, pd.Series):
        y_pred = y_pred.values
    else:
        y_pred = np.array(y_pred)

    return y_true[~np.isnan(y_true) & ~np.isnan(y_pred)], y_pred[~np.isnan(y_true) & ~np.isnan(y_pred)]



import os
import pickle
from typing import List, Union

import pandas as pd

from otb.config import CACHE_DIR


class InMemoryCache:

    def __init__(self, cache_dir: Union[str, os.PathLike, None] = CACHE_DIR) -> None:
        self._cache = dict()
        self._cache_dir = cache_dir

    def add_dataset(self, name: str, dataset: pd.DataFrame) -> None:
        """Adds a new dataset to the cache, persisting to disk if needed."""
        self._cache[name] = dataset
        self._cache_dataset(key=name)

    def available_datasets(self) -> List[str]:
        """List datasets available in memory or on disk"""
        in_memory = self.mem_datasets()
        on_disk = [fp.strip(".pickle") for fp in os.listdir(self._cache_dir) if ".gitkeep" not in str(fp)]
        return list(set(in_memory + on_disk))

    def mem_datasets(self) -> List[str]:
        """List the datasets available in memory."""
        return list(self._cache.keys())

    def get_dataset(self, key: str) -> pd.DataFrame:
        """Get a dataset from memory or disk"""
        if key not in self.available_datasets():
            raise NotImplementedError
        if self._is_in_memory(key):
            return self._cache[key]

    def _is_in_memory(self, key: str) -> bool:
        """Check if a dataset is available in memory."""
        return self.__contains__(key=key)

    def _is_on_disk(self, key: str) -> bool:
        """Check if a dataset was processed and is on disk."""
        on_disk = [fp for fp in os.listdir(self._cache_dir) if ".gitkeep" not in str(fp)]

        for fp in on_disk:
            if fp.split(".pickle")[0] == key:
                return True
        return False

    def _cache_dataset(self, key) -> None:
        """Save a dataset from memory to disk."""
        if key not in self._cache:
            raise KeyError(f"no dataset named {key}.")
        df = self._cache[key]
        df.to_pickle(os.path.join(self._cache_dir, f"{key}.pickle"))

    def _load_dataset(self, key) -> None:
        """Load a dataset from disk to memory"""
        try:
            df = pd.read_pickle(os.path.join(self._cache_dir, f"{key}.pickle"))
            self._cache[key] = df
        except Exception as e:  # @TODO narrow scope
            print(f"failed to load dataset with key '{key}' from cache at {self._cache_dir} with error {e}.")
            return

    def __iter__(self) -> pd.DataFrame:
        i = 0
        keys = list(self._cache.keys())
        while i < len(keys):
            yield self._cache[keys[i]]

    def __len__(self) -> int:
        return len(self._cache)

    def __getitem__(self, key) -> Union[pd.DataFrame, None]:
        if key in self._cache:
            return self._cache[key]
        else:
            return None

    def __contains__(self, key) -> bool:
        return True if key in self._cache else False



from typing import Union

import numpy as np

from otb.benchmark.models.forecasting.base_model import BaseForecastingModel


class LinearForecastingModel(BaseForecastingModel):
    """A model that fits a line to the lagged values of the target variable."""

    def __init__(self, name: str, target_name: str, window_size: int, forecast_horizon: int, **kwargs):
        super().__init__(name, target_name, window_size, forecast_horizon, **kwargs)

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        """Maintain the same interface as the other models."""
        pass

    def predict(self, X: 'pd.DataFrame'):
        """Forecast the cn2 by fitting a line using the lagged values."""
        # obtain the lagged values of the target variable
        X = X[[c for c in X.columns if c.startswith(self.target_name)]]

        # interpolate X to fill in missing values
        X = X.interpolate(method="time")

        # develop a prediction for each row in X
        preds = []
        for i in range(len(X)):
            # fit a line to the lagged values
            lagged_values = X.iloc[i, :].values
            A = np.vstack([np.arange(len(lagged_values)), np.ones(len(lagged_values))]).T
            m, b = np.linalg.lstsq(A, lagged_values, rcond=None)[0]

            # predict the next value at the forecast horizon
            pred = m * (len(lagged_values) + self.forecast_horizon) + b
            preds.append(pred)

        return np.array(preds)



import os
import json
import pprint
from typing import Union

import pandas as pd

from otb.tasks import TaskApi, tasks
from otb.config import BENCHMARK_FP
import otb.benchmark.models.regression as regression_models
import otb.benchmark.models.forecasting as forecasting_models

REGRESSION_MODELS = {n: getattr(regression_models, n) for n in regression_models.__all__}
FORECASTING_MODELS = {n: getattr(forecasting_models, n) for n in forecasting_models.__all__}
PPRINTER = pprint.PrettyPrinter(indent=4, width=120, compact=True)


def run_benchmarks(verbose: bool = True,
                   write_metrics: bool = True,
                   metrics_fp: Union[os.PathLike, str, None] = None) -> dict:
    """Run benchmarks for all tasks and models."""
    if metrics_fp is None:
        metrics_fp = BENCHMARK_FP
    task_api = TaskApi()
    all_tasks = sorted(task_api.list_tasks())
    benchmark_results = {}

    for task_name in all_tasks:
        #if task_name != "regression.mlo_cn2.dropna.Cn2_15m":
        #    continue
        if verbose:
            print(f"Running benchmark for {task_name}...")

        task = task_api.get_task(task_name, benchmark_fp=metrics_fp)
        target_name = task.get_target_name()
        task_info = task.get_info()

        if verbose:
            PPRINTER.pprint(task_info)

        obs_timezone = task_info["obs_tz"]
        obs_lat = task_info["obs_lat"]
        obs_lon = task_info["obs_lon"]
        use_log10 = task_info["log_transform"]

        _, y_test = task.get_test_data(data_type="pd")

        if type(task) == tasks.RegressionTask:
            models = REGRESSION_MODELS
        elif type(task) == tasks.ForecastingTask:
            models = FORECASTING_MODELS
            _, y_test = task.prepare_forecasting_data(_, y_test)
        else:
            raise ValueError(f"unknown task type {type(task)}.")

        benchmark_results[task_name] = {}
        benchmark_results[task_name]["possible_predictions"] = int(y_test.notna().sum().values[0])

        X_train, y_train = task.get_train_data(data_type="pd")
        X_val, y_val = task.get_val_data(data_type="pd")
        X, y = pd.concat([X_train, X_val]), pd.concat([y_train, y_val])
        if type(task) == tasks.ForecastingTask:
            X, y = task.prepare_forecasting_data(X, y)

        if "mlo_cn2" in task_name:
            height_of_observation = 15.0
            air_temperature_col_name = "T_2m"
            water_temperature_col_name = ""
            humidity_col_name = "RH_2m"
            wind_speed_col_name = "Spd_10m"
            time_col_name = "time"
        elif "usna" in task_name:
            if "sm" in task_name:
                air_temperature_col_name = "T_5m"
                wind_speed_col_name = "Spd_10m"
            else:
                air_temperature_col_name = "T_3m"
                wind_speed_col_name = "Spd_3m"
            height_of_observation = 3.0
            water_temperature_col_name = "T_0m"
            humidity_col_name = "RH_3m"
            time_col_name = "time"
        else:
            raise ValueError(f"benchmarks not configured for task {task_name}.")

        for model_name, model in models.items():
            #if model_name != "RandomForestRegressionModel":
            if "AWT" in model_name and "mlo_cn2" in task_name:
                continue
            if verbose:
                print(f"Running benchmark for {model_name}...")

            model_kwargs = dict(name=model_name,
                                target_name=target_name,
                                timezone=obs_timezone,
                                obs_lat=obs_lat,
                                obs_lon=obs_lon,
                                air_temperature_col_name=air_temperature_col_name,
                                water_temperature_col_name=water_temperature_col_name,
                                humidity_col_name=humidity_col_name,
                                wind_speed_col_name=wind_speed_col_name,
                                time_col_name=time_col_name,
                                height_of_observation=height_of_observation,
                                enforce_dynamic_range=True,
                                constant_adjustment=True,
                                use_log10=use_log10)
            # if forecast model, add forecast horizon and window size
            if "forecasting" in task_name:
                model_kwargs["forecast_horizon"] = task.forecast_horizon
                model_kwargs["window_size"] = task.window_size

            mdl = model(**model_kwargs)
            mdl.train(X.copy(deep=True), y.copy(deep=True))  # copy to avoid modifying original data

            results = task.evaluate_model(predict_call=mdl.predict, x_transforms=None, x_transform_kwargs=None)
            benchmark_results[task_name][model_name] = results
            if verbose:
                print(f"Done running benchmark for {model_name}.")
            if verbose:
                PPRINTER.pprint(results)

    if write_metrics:
        with open(metrics_fp, "w") as f:
            f.write(json.dumps(benchmark_results, indent=4))
        if verbose:
            print(f"Wrote benchmark metrics to {metrics_fp}.")
    else:
        if verbose:
            print("Skipping writing benchmark metrics.")

    if verbose:
        print("Done running benchmarks.")
    if verbose:
        PPRINTER.pprint(benchmark_results)
    return benchmark_results



from otb.benchmark.models.forecasting.linear import LinearForecastingModel
from otb.benchmark.models.forecasting.mean_window import MeanWindowForecastingModel
from otb.benchmark.models.forecasting.persistence import PersistenceForecastingModel
from otb.benchmark.models.forecasting.minute_climatology import MinuteClimatologyForecastingModel
from otb.benchmark.models.forecasting.climatology import ClimatologyForecastingModel
from otb.benchmark.models.forecasting.random_forest import RandomForestForecastingModel

__all__ = [
    "LinearForecastingModel", "MeanWindowForecastingModel", "PersistenceForecastingModel",
    "MinuteClimatologyForecastingModel", "ClimatologyForecastingModel"
]



from typing import Union

import pandas as pd
import numpy as np

from otb.benchmark.models.forecasting.base_model import BaseForecastingModel


class ClimatologyForecastingModel(BaseForecastingModel):
    """A model that predicts the mean value of the target variable seen during training."""

    def __init__(self,
                 name: str,
                 target_name: str,
                 window_size: int,
                 forecast_horizon: int,
                 time_col_name: Union[str, None] = None,
                 **kwargs):
        super().__init__(name, target_name, window_size, forecast_horizon, **kwargs)
        self.time_col_name = time_col_name if time_col_name is not None else None
        self.global_mean = np.nan

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        """Determine the mean value of the target variable seen during training."""
        X = X[[c for c in X.columns if c.startswith(self.target_name)]]

        self.global_mean = np.nanmean(X.values.flatten())

    def predict(self, X: 'pd.DataFrame'):
        """Predict the mean seen during training ."""
        return np.full(len(X), self.global_mean)



"""
Offshore Macro-Meteorological Model from:

@article{sadot1992forecasting,
  title={Forecasting optical turbulence strength on the basis of macroscale meteorology and aerosols: models and validation},
  author={Sadot, Dan and Kopeika, Norman S},
  journal={Optical Engineering},
  volume={31},
  number={2},
  pages={200--212},
  year={1992},
  publisher={SPIE}
}
"""
import os
import sys
from typing import Union

import numpy as np
import pandas as pd

from otb.benchmark.models.regression.base_model import BaseRegressionModel
from otb.utils import apply_fried_height_adjustment, add_temporal_hour, add_temporal_hour_weight


class MacroMeteorologicalModel(BaseRegressionModel):
    """A model which predicts the Cn2 value using macro-meteorological parameters."""

    def __init__(self,
                 name: str,
                 target_name: str,
                 timezone: str,
                 obs_lat: float,
                 obs_lon: float,
                 air_temperature_col_name: str,
                 humidity_col_name: str,
                 wind_speed_col_name: str,
                 time_col_name: str,
                 temporal_hour_col_name: str = "temporal_hour",
                 temporal_hour_weight_col_name: str = "temporal_hour_weight",
                 height_of_observation: Union[float, None] = None,
                 enforce_dynamic_range: bool = True,
                 constant_adjustment: bool = True,
                 use_log10: bool = True,
                 **kwargs):
        super().__init__(name, target_name, **kwargs)
        self.timezone = timezone
        self.obs_lat = obs_lat
        self.obs_lon = obs_lon
        self.air_temperature_col_name = air_temperature_col_name
        self.humidity_col_name = humidity_col_name
        self.wind_speed_col_name = wind_speed_col_name
        self.time_col_name = time_col_name
        self.temporal_hour_col_name = temporal_hour_col_name
        self.temporal_hour_weight_col_name = temporal_hour_weight_col_name
        self.height_of_observation = height_of_observation
        self.enforce_dynamic_range = enforce_dynamic_range
        self.constant_adjustment = constant_adjustment
        self.use_log10 = use_log10

    def train(self, X: pd.DataFrame, y: Union[pd.DataFrame, pd.Series, np.ndarray]):
        """Maintain the same interface as the other models."""
        pass

    def predict(self, X: pd.DataFrame):
        """Generate predictions from the model using an input DataFrame."""

        # add temporal hour and temporal hour weight
        if self.temporal_hour_col_name not in X.columns:
            X = add_temporal_hour(X,
                                  name=self.name,
                                  timezone=self.timezone,
                                  obs_lat=self.obs_lat,
                                  obs_lon=self.obs_lon,
                                  time_col_name=self.time_col_name,
                                  temporal_hour_col_name=self.temporal_hour_col_name)
        if self.temporal_hour_weight_col_name not in X.columns:
            X = add_temporal_hour_weight(X,
                                         temporal_hour_col_name=self.temporal_hour_col_name,
                                         temporal_hour_weight_col_name=self.temporal_hour_weight_col_name)

        X_ = X.loc[X[self.air_temperature_col_name].notna() & X[self.humidity_col_name].notna() &
                   X[self.wind_speed_col_name].notna() & X[self.temporal_hour_weight_col_name].notna()].copy()

        T = X_[self.air_temperature_col_name]  # temperature in [C]
        T = T + 273.15  # convert to Kelvin
        U = X_[self.wind_speed_col_name]  # Wind Speed in [m/s]
        RH = X_[self.humidity_col_name]  # Relative Humidity [%]
        W = X_[self.temporal_hour_weight_col_name]  # Temporal Hour Weight

        w = 3.8e-14  # coefficient for temporal hour weight W
        t = 2e-15  # coefficient for temperature (Kelvin) T
        rh = -2.8e-15  # coefficient for relative humidity RH
        rh2 = 2.9e-17  # coefficient for relative humidity squared RH^2
        rh3 = -1.1e-19  # coefficient for relative humidity cubed RH^3
        u = -2.5e-15  # coefficient for wind speed U
        u2 = 1.2e-15  # coefficient for wind speed squared U^2
        u3 = -8.5e-17  # coefficient for wind speed cubed U^3
        c = -5.3e-13  # final coefficient

        mm_prediction = pd.Series((w * W + t * T + rh * RH + rh2 * (RH * RH) + rh3 * (RH * RH * RH) + u * U + u2 *
                                   (U * U) + u3 * (U * U * U) + c),
                                  index=X.index,
                                  name=self.name)

        if self.enforce_dynamic_range:
            X[self.name] = mm_prediction
            X.loc[(X[self.wind_speed_col_name] > 10) | ((X[self.air_temperature_col_name] < 9) |
                                                        (X[self.air_temperature_col_name] > 35)) |
                  ((X[self.humidity_col_name] < 14) | (X[self.humidity_col_name] > 92)), self.name] = np.nan

            mm_prediction = X[self.name].values
            X.drop(columns=[self.name], inplace=True)

        if len(mm_prediction[~np.isnan(mm_prediction)]) > 0:
            if self.constant_adjustment & (min(mm_prediction[~np.isnan(mm_prediction)]) < 0):
                constant_adjustment = min(mm_prediction[~np.isnan(mm_prediction)]) * -1
                mm_prediction = mm_prediction + \
                    constant_adjustment + \
                    sys.float_info.epsilon
            else:
                mm_prediction[mm_prediction <= 0] = np.nan

        # convert from model reference height to height of observation
        if self.height_of_observation is not None:
            mm_prediction = apply_fried_height_adjustment(cn2=mm_prediction,
                                                          observed=15.0,
                                                          desired=self.height_of_observation)

        if self.use_log10:
            return np.log10(mm_prediction)
        return mm_prediction



from typing import Union

import pandas as pd
import numpy as np

from otb.benchmark.models.regression.base_model import BaseRegressionModel


class MinuteClimatologyRegressionModel(BaseRegressionModel):
    """A model that predicts the mean value of the target variable for a given time seen during training."""

    def __init__(self, name: str, target_name: str, time_col_name: Union[str, None] = None, **kwargs):
        super().__init__(name, target_name, **kwargs)
        self.time_col_name = time_col_name if time_col_name is not None else None
        self.means = {}
        self.global_mean = np.nan

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        """Determine the mean value of the target variable seen during training for each time."""
        if isinstance(y, np.ndarray):
            y = pd.Series(y, name=self.target_name, index=X.index)

        self.global_mean = np.nanmean(y.values.flatten())

        # compute the mean for each interval in X across all days
        y["time_of_day"] = y.index.time
        y_means = y.groupby("time_of_day").mean()

        # iterate through the rows in X_means
        for i in range(len(y_means)):
            self.means[y_means.index[i]] = np.nanmean(y_means.iloc[i, :].values)

    def predict(self, X: 'pd.DataFrame'):
        """Predict the mean seen during training at the time of day for inference."""
        times = X.index
        times = pd.to_datetime(times).time

        preds = []

        for time in times:
            if time in self.means and not np.isnan(self.means[time]):
                preds.append(self.means[time])
            else:
                preds.append(self.global_mean)

        return np.array(preds)



from typing import Union

import numpy as np

from otb.benchmark.models.forecasting.base_model import BaseForecastingModel


class MeanWindowForecastingModel(BaseForecastingModel):
    """A model which predicts the mean value of the target variable from the input window."""

    def __init__(self, name: str, target_name: str, window_size: int, forecast_horizon: int, **kwargs):
        super().__init__(name, target_name, window_size, forecast_horizon, **kwargs)

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        """Maintain the same interface as the other models."""
        pass

    def predict(self, X: 'pd.DataFrame'):
        """Forecast the cn2 using the mean of the lagged values."""
        # predict the mean for each entry in X
        # X contains some number of lagged values of the target variable
        # we will use the mean of these lagged values as our prediction
        X = X[[c for c in X.columns if c.startswith(self.target_name)]]

        # develop a prediction for each row in X
        preds = []
        for i in range(len(X)):
            pred = np.nanmean(X.iloc[i, :].values)
            preds.append(pred)

        return np.array(preds)



from typing import Union

import pandas as pd
import numpy as np
import lightgbm as lgb

from otb.benchmark.models.regression.base_model import BaseRegressionModel


class GradientBoostingRegressionModel(BaseRegressionModel):
    """A model that uses a gradient boosting regression tree to predict the target."""

    def __init__(self, name: str, target_name: str, time_col_name: Union[str, None] = None, **kwargs):
        super().__init__(name, target_name, **kwargs)
        self.time_col_name = time_col_name if time_col_name is not None else None
        self._lgb_mdl = lgb.LGBMRegressor(**kwargs)

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        """Train the underlying LightGBM regressor."""
        self._lgb_mdl.fit(X, y)

    def predict(self, X: 'pd.DataFrame'):
        """Use the underlying LightGBM regressor to generate predictions."""
        return self._lgb_mdl.predict(X)



import os
import json
import warnings
from abc import ABC
from datetime import datetime
from enum import Enum
from typing import Any, Callable, List, Tuple, Union

import pandas as pd

import otb.eval.metrics as eval_metrics
from otb.dataset import Dataset
from otb.config import BENCHMARK_FP


class TaskTypes(Enum):
    REGRESSION = "regression"
    FORECASTING = "forecasting"


class TaskABC(ABC):

    def get_info(self, keys: Union[List[str], None] = None) -> dict:
        """Returns the full task information dictionary."""
        raise NotImplementedError

    def get_description(self) -> str:
        """Return the description of the task."""
        raise NotImplementedError

    def get_long_description(self) -> str:
        """Return the description of the task."""
        raise NotImplementedError

    def get_benchmark_info(self, task_name: Union[str, None]) -> dict:
        """Returns the benchmark information dictionary."""
        raise NotImplementedError

    def top_models(self, n: int = 5, metric: str = "") -> List[str]:
        """Returns the top n models for this task."""
        raise NotImplementedError

    def get_transforms(self) -> dict:
        """Return the description of the transforms applied to the X and y data."""
        raise NotImplementedError

    def get_target_name(self) -> str:
        """Return the target feature name for this task."""
        raise NotImplementedError

    def get_unavailable_features(self) -> List[str]:
        """Return the names of features which are unavailable for training and inference in this task."""
        raise NotImplementedError

    def get_metric_names(self) -> List[str]:
        """Return the target feature name for this task."""
        raise NotImplementedError

    def get_dataset(self) -> Dataset:
        """Return the underlying dataset."""
        raise NotImplementedError

    def get_df(self) -> pd.DataFrame:
        """Return the underlying pd.DataFrame for this task's dataset."""
        raise NotImplementedError

    def get_data(self, data_type: str) -> Any:
        """Return the underlying data."""
        raise NotImplementedError

    def get_train_data(self, data_type: str) -> Any:
        """Return the underlying training data for this task."""
        raise NotImplementedError

    def get_test_data(self, data_type: str) -> Any:
        """Return the underlying test data for this task."""
        raise NotImplementedError

    def get_val_data(self, data_type: str) -> Any:
        """Return the underlying validation data for this task."""
        raise NotImplementedError

    def evaluate_model(predict_call: Callable,
                       data_type: str,
                       x_transforms: Union[Callable, None] = None,
                       x_transform_kwargs: Union[dict, None] = None,
                       eval_metric_names: Union[List[str], None] = None,
                       return_predictions: bool = False,
                       include_as_benchmark: bool = False,
                       model_name: Union[str, None] = None,
                       overwrite: bool = False) -> Union[dict, Tuple[dict, 'np.ndarray']]:
        """Evaluate a model against this task's transformed validation set, default against all metrics."""
        raise NotImplementedError


class BaseTask(TaskABC):

    def __init__(self, task_type: str, task_name: str, task: dict, benchmark_fp: str = BENCHMARK_FP) -> None:
        super().__init__()
        self.task_type = task_type
        self.task_name = task_name
        self.task = task
        self.benchmark_fp = benchmark_fp
        self._init_dataset_for_task()

    def get_info(self, keys: Union[List[str], None] = None) -> dict:
        """Returns the task information dictionary."""
        task = self.task
        return {k: task[k] for k in keys} if keys is not None else task

    def get_benchmark_info(self, task_name: Union[str, None] = None) -> dict:
        """Returns the benchmark information dictionary."""
        if os.path.exists(self.benchmark_fp):
            with open(self.benchmark_fp, "r") as f:
                benchmark_info = json.load(f)
            if task_name is None:
                task_name = self.task_name
            if task_name == "*":
                return benchmark_info
            return benchmark_info[task_name]
        else:
            raise FileNotFoundError(
                f"benchmark file {self.benchmark_fp} not found. Please run `otb benchmark` to generate the benchmark file."
            )

    def top_models(self, n: int = 5, metric: str = "") -> List[str]:
        """Returns the top n models for this task."""
        benchmark_info = self.get_benchmark_info(task_name=self.task_name)
        if metric == "":
            metrics = self.get_metric_names()
            metric = metrics[0]
        model_benchmark_info = {k: v for k, v in benchmark_info.items() if k != "possible_predictions" and metric in v}
        model_benchmark_info_keys = sorted(model_benchmark_info.keys(),
                                           key=lambda x: model_benchmark_info[x][metric]["metric_value"])[:n]
        model_benchmark_info = {k: model_benchmark_info[k] for k in model_benchmark_info_keys}
        model_benchmark_info["possible_predictions"] = benchmark_info["possible_predictions"]

        return model_benchmark_info

    def get_description(self) -> str:
        """Return the description of the task."""
        return self.task["description"]

    def get_long_description(self) -> str:
        """Return the description of the task."""
        return self.task["description_long"]

    def get_transforms(self) -> dict:
        """Return the description of the transforms applied to the X and y data."""
        transform_info = {
            "log_transform": {
                "description": "Whether to apply a base-10 log transform to the target.",
                "value": self.task["log_transform"]
            },
            "dropna": {
                "description":
                    "Whether to drop observations with any missing values across both the features and target.",
                "value":
                    self.task["dropna"]
            }
        }

        return transform_info

    def get_target_name(self) -> str:
        """Return the target feature name for this task."""
        return self.task["target"]

    def get_unavailable_features(self) -> List[str]:
        """Return the names of features which are unavailable for training and inference in this task."""
        return self.task["remove"]

    def get_metric_names(self) -> List[str]:
        """Return the target feature name for this task."""
        return self.task["eval_metrics"]

    def get_dataset(self) -> Dataset:
        """Return the underlying dataset."""
        return self._ds

    def get_df(self) -> pd.DataFrame:
        """Return the underlying pd.DataFrame for this task's dataset."""
        return self._ds._df

    def get_data(self, data_type: str = "pd") -> Any:
        """Return the underlying data."""
        return self._ds.get_all(data_type=data_type)

    def get_train_data(self, data_type: str = "pd") -> Any:
        """Return the underlying training data for this task."""
        return self._ds.get_train(task=self.task, data_type=data_type)

    def get_test_data(self, data_type: str = "pd") -> Any:
        """Return the underlying test data for this task."""
        return self._ds.get_test(task=self.task, data_type=data_type)

    def get_val_data(self, data_type: str = "pd") -> Any:
        """Return the underlying validation data for this task."""
        return self._ds.get_val(task=self.task, data_type=data_type)

    def evaluate_model(self,
                       predict_call: Callable,
                       data_type: str = "pd",
                       x_transforms: Union[Callable, None] = None,
                       x_transform_kwargs: Union[dict, None] = None,
                       eval_metric_names: Union[List[str], None] = None,
                       return_predictions: bool = False,
                       include_as_benchmark: bool = False,
                       model_name: Union[str, None] = None,
                       overwrite: bool = True) -> Union[dict, Tuple[dict, 'np.ndarray']]:
        """Evaluate a model against this task's transformed test set, default against all metrics."""
        raise NotImplementedError

    def _init_dataset_for_task(self) -> None:
        """Get the underlying dataset for the given task"""
        self._ds = Dataset(name=self.task["ds_name"])

    def _add_experiment_to_benchmarks(self, model_name: str, model_metrics: dict, overwrite: bool) -> None:
        if model_name is None:
            raise ValueError(f"model_name must be provided if include_as_benchmark is True.")
        benchmark_info = self.get_benchmark_info(task_name="*")
        if not overwrite and model_name in benchmark_info:
            model_name = f"{model_name}_{datetime.now().strftime('%Y%m%d%H%M%S')}"
        benchmark_info[self.task_name][model_name] = model_metrics
        with open(self.benchmark_fp, "w") as f:
            json.dump(benchmark_info, f, indent=4)


class RegressionTask(BaseTask):

    def __init__(self, task_type: str, task_name: str, task: dict, benchmark_fp: str = BENCHMARK_FP) -> None:
        super().__init__(task_type=task_type, task_name=task_name, task=task, benchmark_fp=benchmark_fp)

    def evaluate_model(self,
                       predict_call: Callable,
                       predict_call_kwargs: Union[dict, None] = None,
                       data_type: str = "pd",
                       x_transforms: Union[Callable, None] = None,
                       x_transform_kwargs: Union[dict, None] = None,
                       eval_metric_names: Union[List[str], None] = None,
                       return_predictions: bool = False,
                       include_as_benchmark: bool = False,
                       model_name: Union[str, None] = None,
                       overwrite: bool = True) -> Union[dict, Tuple[dict, 'np.ndarray']]:
        """Evaluate a model against this task's transformed test set, default against all metrics."""
        # obtain evaluation data
        X_test, y_test = self.get_test_data(data_type=data_type)

        # apply x_transforms if present
        if x_transforms is not None:
            if x_transform_kwargs is not None:
                X_test = x_transforms(X_test, **x_transform_kwargs)
            else:
                X_test = x_transforms(X_test)

        # get predictions
        if predict_call_kwargs is not None:
            y_test_pred = predict_call(X_test, **predict_call_kwargs)
        else:
            y_test_pred = predict_call(X_test)

        if eval_metric_names is None:
            eval_metric_names = self.task["eval_metrics"]
        model_metrics = {k: -1 for k in eval_metric_names}

        for m in eval_metric_names:
            val = getattr(eval_metrics, m)(y_test, y_test_pred)
            model_metrics[m] = val

        if include_as_benchmark:
            self._add_experiment_to_benchmarks(model_name=model_name, model_metrics=model_metrics, overwrite=overwrite)
        if return_predictions:
            return model_metrics, y_test_pred
        return model_metrics


class ForecastingTask(BaseTask):

    def __init__(self, task_type: str, task_name: str, task: dict, benchmark_fp: str = BENCHMARK_FP) -> None:
        super().__init__(task_type=task_type, task_name=task_name, task=task, benchmark_fp=benchmark_fp)
        self.forecast_horizon = self.task["forecast_horizon"]
        self.window_size = self.task["window_size"]

    def prepare_forecasting_data(self,
                                 X: pd.DataFrame,
                                 y: Union[pd.DataFrame, pd.Series],
                                 window_size: Union[int, None] = None,
                                 forecast_horizon: Union[int, None] = None):
        """Prepare data for forecasting."""
        window_size = window_size if window_size is not None else self.window_size
        forecast_horizon = forecast_horizon if forecast_horizon is not None else self.forecast_horizon

        X = self._join_target(X, y)
        X = self._add_lags(X, window_size)
        y = self._shift_target(y, forecast_horizon)
        X, y = self._obtain_valid_data(X, y, window_size, forecast_horizon)

        return X, y

    def evaluate_model(self,
                       predict_call: Callable,
                       predict_call_kwargs: Union[dict, None] = None,
                       window_size: Union[int, None] = None,
                       forecast_horizon: Union[int, None] = None,
                       data_type: str = "pd",
                       x_transforms: Union[Callable, None] = None,
                       x_transform_kwargs: Union[dict, None] = None,
                       forecast_transforms: Union[Callable, None] = None,
                       forecast_transform_kwargs: Union[dict, None] = None,
                       eval_metric_names: Union[List[str], None] = None,
                       return_predictions: bool = False,
                       include_as_benchmark: bool = False,
                       model_name: Union[str, None] = None,
                       overwrite: bool = True) -> Union[dict, Tuple[dict, 'np.ndarray']]:
        """Evaluate a model against this task's transformed test set, default against all metrics."""
        window_size = window_size if window_size is not None else self.window_size
        forecast_horizon = forecast_horizon if forecast_horizon is not None else self.forecast_horizon

        # obtain evaluation data
        X_test, y_test = self.get_test_data(data_type=data_type)
        assert forecast_horizon + window_size < len(
            X_test
        ), f"window_size and forecast_horizon must be less than the length of the evaluation set ({len(X_test)})."

        # apply x_transforms if present
        if x_transforms is not None:
            if x_transform_kwargs is not None:
                X_test = x_transforms(X_test, **x_transform_kwargs)
            else:
                X_test = x_transforms(X_test)

        # apply window and forecasting horizon transforms
        if forecast_transforms is not None:
            if forecast_transform_kwargs is not None:
                y_test = forecast_transforms(y_test, **forecast_transform_kwargs)
            else:
                y_test = forecast_transforms(y_test)

        # obtain forecasting data
        X_test, y_test = self.prepare_forecasting_data(X_test, y_test, window_size, forecast_horizon)

        # get predictions
        if predict_call_kwargs is not None:
            y_test_pred = predict_call(X_test, **predict_call_kwargs)
        else:
            y_test_pred = predict_call(X_test)

        if eval_metric_names is None:
            eval_metric_names = self.task["eval_metrics"]
        model_metrics = {k: -1 for k in eval_metric_names}

        for m in eval_metric_names:
            val = getattr(eval_metrics, m)(y_test, y_test_pred)
            model_metrics[m] = val

        if include_as_benchmark:
            self._add_experiment_to_benchmarks(model_name=model_name, model_metrics=model_metrics, overwrite=overwrite)
        if return_predictions:
            return model_metrics, y_test_pred
        return model_metrics

    def _join_target(self, X, y):
        """Include the target in the features."""
        X = X.join(y)

        return X

    def _add_lags(self, X, window_size):
        """Lag all feature columns from 1 to lags inclusive."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", pd.errors.PerformanceWarning)
            X = X.assign(
                **{f'{col} (t-{lag})': X[col].shift(lag) for lag in range(1, window_size + 1) for col in X.columns})

        return X

    def _shift_target(self, y, forecast_horizon):
        """Shift the target by the forecast horizon."""
        y = y.shift(-forecast_horizon)

        return y

    def _obtain_valid_data(self, X, y, window_size, forecast_horizon):
        """Obtain data that is valid for training and evaluation."""
        X = X[window_size:-1 * forecast_horizon]
        y = y[window_size:-1 * forecast_horizon]

        return X, y


class TaskApi(object):
    """A factory for creating optical turbulence modeling tasks."""

    def __init__(self, root_dir: Union[str, None] = None) -> None:
        """Read the currently-supported benchmarking task for loaders and evaluators."""
        if root_dir is None:
            root_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))  # obtain root path
        tasks_path = os.path.join(root_dir, 'config', 'tasks.json')
        tasks = json.load(open(tasks_path, 'rb'))

        self.tasks = tasks
        self._build_task_names()

    def get_task(self, task_name: str, benchmark_fp: str = BENCHMARK_FP) -> Union[RegressionTask, ForecastingTask]:
        """Get a task by name."""
        if self._is_supported_task(task_name=task_name):
            if task_name.split(".")[0] == TaskTypes.REGRESSION.value:
                return RegressionTask(task_type=TaskTypes.REGRESSION,
                                      task_name=task_name,
                                      task=self._get_task(key=task_name),
                                      benchmark_fp=benchmark_fp)
            elif task_name.split(".")[0] == TaskTypes.FORECASTING.value:
                return ForecastingTask(task_type=TaskTypes.REGRESSION,
                                       task_name=task_name,
                                       task=self._get_task(key=task_name),
                                       benchmark_fp=benchmark_fp)
            else:
                raise NotImplementedError(f"task type {task_name.split('.')[0]} not supported.")
        else:
            raise NotImplementedError(f"task {task_name} not supported.")

    def list_tasks(self) -> List[str]:
        """List all currently supported tasks."""
        return list(self.task_names)

    def _get_task(self, key: str) -> dict:
        """Traverse a task key, returning bottom-level data."""
        path = key.split(".")
        d = self.tasks.copy()
        for i in range(len(path)):
            d = d[path[i]]
        return d

    def _is_supported_task(self, task_name: str) -> bool:
        """Check whether a task is supported by otbench."""
        return task_name in self.task_names

    def _build_task_names(self) -> None:
        """Gather `.` separated task names for easier inclusion checking."""
        task_names = set()
        task_paths = [(t, t, v) for t, v in self.tasks.items()]
        next_paths = []

        while task_paths or next_paths:
            if not task_paths:
                task_paths = next_paths
                next_paths = []

            _, full_k, v = task_paths.pop()

            if 'description' in v:
                task_names.add(full_k)
            else:
                next_paths.extend([(nk, f'{full_k}.{nk}', nv) for nk, nv in v.items()])

        self.task_names = task_names



import os

CONFIG_DIR = os.path.dirname(os.path.abspath(__file__))
TASKS_FP = os.path.join(CONFIG_DIR, "tasks.json")
DATASETS_FP = os.path.join(CONFIG_DIR, "datasets.json")
ROOT_DIR = os.path.dirname(CONFIG_DIR)
CACHE_DIR = os.path.join(ROOT_DIR, "cache", "processed")
DATA_DIR = os.path.join(ROOT_DIR, "data")
BENCHMARK_FP = os.path.join(ROOT_DIR, "benchmark", "experiments.json")
RETURN_TYPES = ["pd", "np", "xr", "nc"]



from typing import Union

import numpy as np

from otb.benchmark.models.regression.base_model import BaseRegressionModel


class PersistenceRegressionModel(BaseRegressionModel):
    """A model which predicts the most recent value of the target variable."""

    def __init__(self, name: str, target_name: str, **kwargs):
        super().__init__(name, target_name, **kwargs)
        self.persistence = np.nan

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        """Maintain the same interface as the other models."""
        self.persistence = y.values[-1]

    def predict(self, X: 'pd.DataFrame'):
        # predict the mean for each entry in X
        return np.full(len(X), self.persistence)



from typing import Union

import pandas as pd
import numpy as np
import lightgbm as lgb

from otb.benchmark.models.forecasting.base_model import BaseForecastingModel


class RandomForestForecastingModel(BaseForecastingModel):
    """A model that uses a random forest regressor for direct forecasting."""

    def __init__(self, name: str, target_name: str, window_size: int, forecast_horizon: int, **kwargs):
        super().__init__(name, target_name, window_size, forecast_horizon, **kwargs)
        self._lgb_mdl = lgb.LGBMRegressor(**kwargs)

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        """Train the underlying LightGBM regressor."""
        self._lgb_mdl.fit(X, y)

    def predict(self, X: 'pd.DataFrame'):
        """Use the underlying LightGBM regressor to generate predictions."""
        return self._lgb_mdl.predict(X)



from typing import Union

import numpy as np

from otb.benchmark.models.forecasting.base_model import BaseForecastingModel


class PersistenceForecastingModel(BaseForecastingModel):
    """A model which predicts the most recent value of the target variable."""

    def __init__(self, name: str, target_name: str, window_size: int, forecast_horizon: int, **kwargs):
        super().__init__(name, target_name, window_size, forecast_horizon, **kwargs)

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        """Maintain the same interface as the other models."""
        pass

    def predict(self, X: 'pd.DataFrame'):
        persistence = X[self.target_name].values[0]

        # develop a prediction for each row in X
        return np.array([persistence for i in range(len(X))])



import warnings

import pandas as pd
from astral import LocationInfo
from astral.sun import sun

warnings.filterwarnings("ignore", category=pd.errors.PerformanceWarning)


def add_temporal_hour(X: pd.DataFrame, name, timezone, obs_lat, obs_lon, time_col_name, temporal_hour_col_name):
    """Get the difference between the current time and sunrise."""

    OBS = LocationInfo(name=name, region='', timezone=timezone, latitude=obs_lat, longitude=obs_lon)

    def _get_sunrise_sunset(dt):
        s = sun(OBS.observer, date=dt, tzinfo=OBS.timezone)
        return (s['sunrise'], s['sunset'])

    if isinstance(X.index, pd.DatetimeIndex):
        X['date'] = X.index.date
        if time_col_name not in X.columns:
            X[time_col_name] = X.index
    else:
        X['date'] = pd.to_datetime(X[time_col_name]).dt.date

    suntimes = X['date'].apply(lambda x: _get_sunrise_sunset(x))
    X['time_sunrise'] = suntimes.apply(lambda x: x[0])
    X['time_sunset'] = suntimes.apply(lambda x: x[1])

    X['time_sunrise'] = X['time_sunrise'].dt.tz_convert(timezone)
    X['time_sunrise'] = X['time_sunrise'].dt.tz_localize(tz=None)
    X['time_sunset'] = X['time_sunset'].dt.tz_convert(timezone)
    X['time_sunset'] = X['time_sunset'].dt.tz_localize(tz=None)

    X[temporal_hour_col_name] = ((X[time_col_name] - X['time_sunrise']).dt.total_seconds()) / (
        (1 / 12) * (X[time_col_name] - X['time_sunrise']).dt.total_seconds())

    X.drop(columns=['date', 'time_sunrise', 'time_sunset'], inplace=True)

    return X


def add_temporal_hour_weight(X: pd.DataFrame, temporal_hour_col_name: str, temporal_hour_weight_col_name: str):
    """Get the temporal hour weight from the temporal hour"""

    TEMPORAL_HOUR_DICT = {
        -99: 0.11,
        -4: 0.11,
        -3: 0.07,
        -2: 0.08,
        -1: 0.06,
        0: 0.05,
        1: 0.10,
        2: 0.51,
        3: 0.75,
        4: 0.95,
        5: 1,
        6: 0.9,
        7: 0.8,
        8: 0.59,
        9: 0.32,
        10: 0.22,
        11: 0.10,
        12: 0.08,
        13: 0.13,
        99: 0.13
    }

    X[temporal_hour_weight_col_name] = 0
    keys = list(TEMPORAL_HOUR_DICT.keys())

    for idx in range(len(keys) - 1):
        # read values (start, stop, W_th) from the dictionary
        start_temporal_hour = keys[idx]
        end_temporal_hour = keys[idx + 1]
        temporal_hour_weight_value = TEMPORAL_HOUR_DICT[keys[idx]]

        # assign DataFrame values based on temporal hour of observation and bin in dict
        X.loc[(X[temporal_hour_col_name] > start_temporal_hour) & (X[temporal_hour_col_name] < end_temporal_hour),
              temporal_hour_weight_col_name] = temporal_hour_weight_value

    return X



"""
Offshore Macro-Meteorological Model from:

@article{wang2015prediction,
  title={Prediction model of atmospheric refractive index structure parameter in coastal area},
  author={Wang, Hongxing and Li, Bifeng and Wu, Xiaojun and Liu, Chuanhui and Hu, Zhihui and Xu, Pengfei},
  journal={Journal of Modern Optics},
  volume={62},
  number={16},
  pages={1336--1346},
  year={2015},
  publisher={Taylor \& Francis}
}
"""
import sys
from typing import Union

import numpy as np
import pandas as pd

from otb.benchmark.models.regression.base_model import BaseRegressionModel
from otb.utils import apply_fried_height_adjustment, add_temporal_hour, add_temporal_hour_weight


class OffshoreMacroMeteorologicalModel(BaseRegressionModel):
    """A model which predicts the Cn2 value using macro-meteorological parameters, tuned for offshore cases."""

    def __init__(self,
                 name: str,
                 target_name: str,
                 timezone: str,
                 obs_lat: float,
                 obs_lon: float,
                 air_temperature_col_name: str,
                 humidity_col_name: str,
                 wind_speed_col_name: str,
                 time_col_name: str,
                 temporal_hour_col_name: str = "temporal_hour",
                 temporal_hour_weight_col_name: str = "temporal_hour_weight",
                 height_of_observation: Union[float, None] = None,
                 enforce_dynamic_range: bool = True,
                 constant_adjustment: bool = True,
                 use_log10: bool = True,
                 **kwargs):
        super().__init__(name, target_name, **kwargs)
        self.timezone = timezone
        self.obs_lat = obs_lat
        self.obs_lon = obs_lon
        self.air_temperature_col_name = air_temperature_col_name
        self.humidity_col_name = humidity_col_name
        self.wind_speed_col_name = wind_speed_col_name
        self.time_col_name = time_col_name
        self.temporal_hour_col_name = temporal_hour_col_name
        self.temporal_hour_weight_col_name = temporal_hour_weight_col_name
        self.height_of_observation = height_of_observation
        self.enforce_dynamic_range = enforce_dynamic_range
        self.constant_adjustment = constant_adjustment
        self.use_log10 = use_log10

    def train(self, X: pd.DataFrame, y: Union[pd.DataFrame, pd.Series, np.ndarray]):
        """Maintain the same interface as the other models."""
        pass

    def predict(self, X: pd.DataFrame):
        """Generate predictions from the model using an input DataFrame."""

        # add temporal hour and temporal hour weight
        if self.temporal_hour_col_name not in X.columns:
            X = add_temporal_hour(X,
                                  name=self.name,
                                  timezone=self.timezone,
                                  obs_lat=self.obs_lat,
                                  obs_lon=self.obs_lon,
                                  time_col_name=self.time_col_name,
                                  temporal_hour_col_name=self.temporal_hour_col_name)
        if self.temporal_hour_weight_col_name not in X.columns:
            X = add_temporal_hour_weight(X,
                                         temporal_hour_col_name=self.temporal_hour_col_name,
                                         temporal_hour_weight_col_name=self.temporal_hour_weight_col_name)

        X_ = X.loc[X[self.air_temperature_col_name].notna() & X[self.humidity_col_name].notna() &
                   X[self.wind_speed_col_name].notna() & X[self.temporal_hour_weight_col_name].notna()].copy()

        T = X_[self.air_temperature_col_name]  # temperature in [C]
        U = X_[self.wind_speed_col_name]  # Wind Speed in [m/s]
        RH = X_[self.humidity_col_name]  # Relative Humidity [%]
        W = X_[self.temporal_hour_weight_col_name]  # Temporal Hour Weight

        w = -1.58e-15  # coefficient for temporal hour weight W
        t = 2.74e-16  # coefficient for temperature T
        rh = 8.3e-17  # coefficient for relative humidity RH
        rh2 = -2.22e-18  # coefficient for relative humidity squared RH^2
        rh3 = 1.42e-20  # coefficient for relative humidity cubed RH^3
        u = 3.37e-16  # coefficient for wind speed U
        u2 = 1.92e-16  # coefficient for wind speed squared U^2
        u3 = -2.8e-17  # coefficient for wind speed cubed U^3
        c = -7.44e-14  # final coefficient

        omm_prediction = pd.Series((w * W + t * T + rh * RH + rh2 * (RH * RH) + rh3 * (RH * RH * RH) + u * U + u2 *
                                    (U * U) + u3 * (U * U * U) + c),
                                   index=X.index,
                                   name=self.name)

        if self.enforce_dynamic_range:
            X[self.name] = omm_prediction
            X.loc[((X[self.wind_speed_col_name] > 15) | ((X[self.air_temperature_col_name] < -2) |
                                                         (X[self.air_temperature_col_name] > 24)) |
                   (X[self.humidity_col_name] < 15)), self.name] = np.nan
            omm_prediction = X[self.name].values

            X.drop(columns=[self.name], inplace=True)

        if len(omm_prediction[~np.isnan(omm_prediction)]) > 0:
            if self.constant_adjustment & (min(omm_prediction[~np.isnan(omm_prediction)]) <= 0):
                constant_adjustment = min(omm_prediction[~np.isnan(omm_prediction)]) * -1

                omm_prediction = (
                    omm_prediction + \
                    constant_adjustment + \
                    sys.float_info.epsilon  # add epsilon to prevent 0 values
                    )
            else:
                omm_prediction[omm_prediction <= 0] = np.nan

        # convert from model reference height to height of observation
        if self.height_of_observation is not None:
            omm_prediction = apply_fried_height_adjustment(cn2=omm_prediction,
                                                           observed=62.0,
                                                           desired=self.height_of_observation)

        if self.use_log10:
            return np.log10(omm_prediction)
        return omm_prediction



import setuptools

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name='otbench',
    packages=setuptools.find_packages(exclude=['*tests*', '*notebooks*']),
    version='0.23.10.26',
    license='MIT',
    description='Consistent benchmarks for evaluating optical turbulence strength models.',
    author='Chris Jellen',
    author_email='cdjellen@gmail.com',
    url='https://github.com/cdjellen/otbench',
    long_description=long_description,
    long_description_content_type='text/markdown',
    download_url='https://github.com/cdjellen/otbench/tarball/main',
    keywords=['optics', 'turbulence', 'benchmark', 'python3'],
    install_requires=['astral', 'pandas', 'requests', 'scikit-learn', 'xarray'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)



from typing import Union

import pandas as pd
import numpy as np

from otb.benchmark.models.regression.base_model import BaseRegressionModel


class ClimatologyRegressionModel(BaseRegressionModel):
    """A model that predicts the mean value of the target variable seen during training."""

    def __init__(self, name: str, target_name: str, time_col_name: Union[str, None] = None, **kwargs):
        super().__init__(name, target_name, **kwargs)
        self.time_col_name = time_col_name if time_col_name is not None else None
        self.global_mean = np.nan

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        """Determine the mean value of the target variable seen during training."""
        if isinstance(y, np.ndarray):
            y = pd.Series(y, name=self.target_name, index=X.index)

        self.global_mean = np.nanmean(y[self.target_name].values.flatten())

    def predict(self, X: 'pd.DataFrame'):
        """Predict the mean seen during training at the time of day for inference."""
        return np.full(len(X), self.global_mean)



from otb.benchmark.models.regression.air_water_temperature_difference import AWTModel
from otb.benchmark.models.regression.macro_meteorological import MacroMeteorologicalModel
from otb.benchmark.models.regression.offshore_macro_meteorological import OffshoreMacroMeteorologicalModel
from otb.benchmark.models.regression.persistence import PersistenceRegressionModel
from otb.benchmark.models.regression.minute_climatology import MinuteClimatologyRegressionModel
from otb.benchmark.models.regression.climatology import ClimatologyRegressionModel
from otb.benchmark.models.regression.hybrid_awt import HybridAWTRegressionModel
from otb.benchmark.models.regression.gradient_boosting_regression_tree import GradientBoostingRegressionModel

__all__ = [
    "AWTModel", "MacroMeteorologicalModel", "OffshoreMacroMeteorologicalModel", "PersistenceRegressionModel",
    "MinuteClimatologyRegressionModel", "ClimatologyRegressionModel", "HybridAWTRegressionModel",
    "GradientBoostingRegressionModel"
]



"""
Air-Water Temperature Difference Model from:

@article{chen2019climatological,
  title={Climatological analysis of the seeing at Fuxian Solar Observatory},
  author={Chen, Li-Hui and Liu, Zhong and Chen, Dong},
  journal={Research in Astronomy and Astrophysics},
  volume={19},
  number={1},
  pages={015},
  year={2019},
  publisher={IOP Publishing}
}
"""
import numpy as np
import pandas as pd

from otb.benchmark.models.regression.base_model import BaseRegressionModel


class AWTModel(BaseRegressionModel):

    def __init__(self,
                 name: str,
                 target_name: str,
                 air_temperature_col_name: str,
                 water_temperature_col_name: str,
                 use_log10: bool = True,
                 **kwargs):
        super().__init__(name, target_name, **kwargs)
        self.air_temperature_col_name = air_temperature_col_name
        self.water_temperature_col_name = water_temperature_col_name
        self.use_log10 = use_log10

    def train(self, X: 'pd.DataFrame', y: 'pd.DataFrame'):
        """Maintain the same interface as the other models."""
        pass

    def predict(self, X):
        """Predict the log10 of the cn2 value using the air-water temperature difference."""
        X_ = X.loc[X[self.air_temperature_col_name].notna() & X[self.water_temperature_col_name].notna()].copy()

        awt = X_[self.air_temperature_col_name] - X_[self.water_temperature_col_name]
        awt_prediction = pd.Series((2.05 * awt**2 + 2.37 * awt + 1.58) * 1e-16, index=X.index, name=self.name)

        if self.use_log10:
            return np.log10(awt_prediction)
        return awt_prediction



from typing import Union

import pandas as pd
import numpy as np


class BaseForecastingModel:
    """A model that predicts the mean value of the target variable for a given time."""

    def __init__(self,
                 name: str,
                 target_name: str,
                 window_size: int,
                 forecast_horizon: int,
                 time_col_name: Union[str, None] = None,
                 **kwargs):
        self.name = name
        self.target_name = target_name
        self.window_size = window_size
        self.forecast_horizon = forecast_horizon

    def train(self, X: 'pd.DataFrame', y: Union['pd.DataFrame', 'pd.Series', np.ndarray]):
        # use observations of the past `window_size` to predict the strength of optical turbulence at the forecast horizon
        raise NotImplementedError

    def predict(self, X: 'pd.DataFrame'):
        # predict the strength of optical turbulence at the forecast horizon
        raise NotImplementedError
